"""Extra information about the reflection API
"""
import inspect
from typing import Dict, Type, List, Callable

RETURN_ANNOTATION = "return"


class FunctionSpec:
    """
    Describes the arguments of a function
    """

    args: List
    annotations: Dict[str, Type]
    return_type: Type

    def __init__(self, args, annotations, return_type):
        self.args = args
        self.annotations = annotations
        self.return_type = return_type


class CachingReflector:
    """
    Takes a function and returns an object representing
    the function's type signature. Results are cached
    so subsequent calls do not need to call the reflection
    API.
    """

    _reflection_cache: Dict[Callable, FunctionSpec]

    def __init__(self):
        self._reflection_cache = {}

    def get_function_spec(self, func) -> FunctionSpec:
        """
        Returns details about the function's signature
        :param func:
        :return:
        """
        if func not in self._reflection_cache:
            spec = inspect.getfullargspec(func)
            ret = spec.annotations.pop(RETURN_ANNOTATION, None)
            self._reflection_cache[func] = FunctionSpec(
                spec.args, spec.annotations, ret
            )
        return self._reflection_cache[func]



from lagom import Container, Singleton, bind_to_container
import time


class SomeService:
    def do_it(self):
        pass


class AThingIMightNeed:
    service: SomeService

    def __init__(self, service: SomeService):
        self.service = service

    def do_it(self):
        self.service.do_it()


container = Container()
container[SomeService] = Singleton(SomeService)


@bind_to_container(container)
def do_work(thing: AThingIMightNeed):
    thing.do_it()


if __name__ == "__main__":
    start = time.perf_counter()
    for i in range(0, 100_000):
        do_work()
    end = time.perf_counter()
    print(f"Time taken {end - start}")

# 3.5875363860000107



import re, textwrap
from sybil import Region, Sybil

FENCE_START = re.compile(r'```python')
FENCE_END = re.compile(r'```')


def parse_bash_blocks(document):
    for start_match, end_match, source in document.find_region_sources(
            FENCE_START, FENCE_END
    ):
        yield Region(start_match.start(), end_match.end(), source, evaluate_code_block)


def evaluate_code_block(example):
    code = compile(example.parsed, example.document.path, 'exec', dont_inherit=True)
    exec(code, example.namespace)
    # exec adds __builtins__, we don't want it:
    del example.namespace['__builtins__']


pytest_collect_file = Sybil(
    parsers=[parse_bash_blocks],
    pattern='cookbook.md', ).pytest()



"""
This module provides code to automatically load environment variables from the container.
It is built on top of (and requires) pydantic.

At first one or more classes representing the required environment variables are defined.
All environment variables are assumed to be all uppercase and are automatically lowercased.

class MyWebEnv(Env):
    port: str
    host: str

class DBEnv(Env):
    db_host: str
    db_password: str


@bind_to_container(c)
def some_function(env: DBEnv):
    do_something(env.db_host, env.db_password)

"""
import os
from abc import ABC

from pydantic.main import BaseModel


class Env(ABC, BaseModel):
    """
    This class implements logic to automatically load properties from ENV
    variables.
    """

    def __init__(self, **kwargs):
        if len(kwargs) == 0:
            super().__init__(
                **{key.lower(): value for (key, value) in os.environ.items()}
            )
        else:
            super().__init__(**kwargs)



from dataclasses import dataclass
from typing import Generator, Any

import pytest

from lagom import Container
from lagom.decorators import dependency_definition
from lagom.exceptions import MissingReturnType

global finally_was_executed
finally_was_executed = False


@dataclass
class MyComplexDep:
    some_number: int


@dataclass
class WrapperOfSomeKind:
    inner: MyComplexDep


def test_functions_that_are_typed_can_be_used_by_a_container(container: Container):
    @dependency_definition(container)
    def my_constructor() -> MyComplexDep:
        return MyComplexDep(some_number=5)

    assert container[MyComplexDep] == MyComplexDep(some_number=5)


def test_the_functions_return_new_instances_each_time(container: Container):
    @dependency_definition(container)
    def my_constructor() -> MyComplexDep:
        return MyComplexDep(some_number=5)

    first = container[MyComplexDep]
    second = container[MyComplexDep]
    assert first is not second


def test_definition_functions_can_yield_instead_of_returning(container: Container):
    @dependency_definition(container)
    def my_constructor() -> Generator[MyComplexDep, Any, Any]:
        yield MyComplexDep(some_number=5)

    first = container[MyComplexDep]
    assert first.some_number == 5


def test_when_yielding_finally_can_be_used(container: Container):
    global finally_was_executed
    finally_was_executed = False

    @dependency_definition(container)
    def my_constructor() -> Generator[MyComplexDep, Any, Any]:
        global finally_was_executed
        try:
            yield MyComplexDep(some_number=5)
        finally:
            finally_was_executed = True

    container.resolve(MyComplexDep)
    assert finally_was_executed is True


def test_functions_can_be_made_into_singletons(container: Container):
    @dependency_definition(container, singleton=True)
    def my_constructor() -> MyComplexDep:
        return MyComplexDep(some_number=5)

    first = container[MyComplexDep]
    second = container[MyComplexDep]
    assert first is second


def test_definition_functions_get_an_instance_of_the_container(container: Container):
    container[MyComplexDep] = MyComplexDep(some_number=3)

    @dependency_definition(container)
    def my_constructor(c: Container) -> WrapperOfSomeKind:
        return WrapperOfSomeKind(c[MyComplexDep])

    assert container[WrapperOfSomeKind].inner == container[MyComplexDep]


def test_singleton_definition_functions_get_an_instance_of_the_container(
    container: Container,
):
    container[MyComplexDep] = MyComplexDep(some_number=3)

    @dependency_definition(container, singleton=True)
    def my_constructor(c: Container) -> WrapperOfSomeKind:
        return WrapperOfSomeKind(c[MyComplexDep])

    assert container[WrapperOfSomeKind].inner == container[MyComplexDep]


def test_functions_that_are_not_typed_raise_an_error(container: Container):
    with pytest.raises(MissingReturnType):

        @dependency_definition(container)
        def my_constructor():
            return MyComplexDep(some_number=5)



from typing import Any

import pytest

from lagom.experimental.integrations.django import DjangoContainer


class FakeDjangoManager:
    pass


class FakeDjangoModel:
    data: Any
    objects = FakeDjangoManager()
    custom_manager = FakeDjangoManager()

    def __init__(self, **kwargs):
        self.data = kwargs


@pytest.fixture(scope="function")
def django_container():
    return DjangoContainer(models=[FakeDjangoModel])  # type: ignore



import pytest

from lagom import Container
from lagom.experimental.integrations.django import DjangoModel
from tests.experimental.integrations.conftest import FakeDjangoModel, FakeDjangoManager


def test_django_models_can_be_created(django_container: Container):
    models = django_container.resolve(DjangoModel[FakeDjangoModel])
    new_model_instance = models.new(first=1, second=2)
    assert isinstance(new_model_instance, FakeDjangoModel)
    assert new_model_instance.data == {"first": 1, "second": 2}


def test_django_models_can_be_managed(django_container: Container):
    models = django_container.resolve(DjangoModel[FakeDjangoModel])
    assert isinstance(models.objects, FakeDjangoManager)


def test_django_models_can_be_managed_via_a_custom_manager(django_container: Container):
    models = django_container.resolve(DjangoModel[FakeDjangoModel])
    assert isinstance(models.managers.custom_manager, FakeDjangoManager)


def test_trying_to_load_a_non_existent_manager_throws_a_sensible_error(
    django_container: Container,
):
    models = django_container.resolve(DjangoModel[FakeDjangoModel])
    with pytest.raises(KeyError):
        models.managers.lol.all()



from typing import Optional

from lagom import Container
from lagom.experimental.environment import Env


class MyEnv(Env):
    example_value: Optional[str]


def test_env_can_be_loaded(container: Container):
    assert isinstance(container.resolve(MyEnv), MyEnv)


def test_defined_env_variables_are_loaded_automatically(
    monkeypatch, container: Container
):
    monkeypatch.setenv("EXAMPLE_VALUE", "from the environment")
    env = container.resolve(MyEnv)
    assert env.example_value == "from the environment"
